/*
 * Copyright (c) 2006 Ed Schouten <ed@fxq.nl>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */
/**
 * @file gui/gui_browser.c
 */

#include <string.h>

#include <config.h>
#include <gui/gui_internal.h>
#include <playq.h>
#include <vfs/vfs.h>
#include <trans.h>

static struct vfs_ent *ve_curdir, *ve_top, *ve_selected;
static WINDOW *win_dirname, *win_browser;

static void
gui_browser_cursor_adjust_top(void)
{
	int i;
	struct vfs_ent *ve;

	for (ve = ve_top, i = 1;
	    (ve != NULL) && (i < gui_size_browser_height); i++) {
		ve = vfs_population_next(ve);

		/* The selected item is on screen */
		if (ve == ve_selected)
			return;
	}

	/* The selection is not on screen */
	ve_top = ve_selected;
}

/**
 * @brief Align the viewport that the selected object is at most at the
 *        bottom of the list of files.
 */
static void
gui_browser_cursor_adjust_bottom(void)
{
	int i;
	struct vfs_ent *ve;

	ve = ve_selected;

	/* Don't deal with empty selections */
	if (ve == NULL)
		return;
	
	for (i = 1; i < gui_size_browser_height; i++) {
		ve = vfs_population_prev(ve);

		/* The top of the list is in screen */
		if ((ve == NULL) || (ve == ve_top))
			return;
	}

	/* We've gone too far */
	ve_top = ve;
}

/**
 * @brief Align the viewport that the selected object resides in the
 *        middle of the view.
 */
static void
gui_browser_cursor_adjust_middle(void)
{
	int i;

	ve_top = ve_selected;

	/* Don't deal with empty directories */
	if (ve_top == NULL)
		return;

	for (i = 0; i < (gui_size_browser_height - 1) / 2; i++) {
		ve_top = vfs_population_prev(ve_top);

		if (ve_top == NULL) {
			/* We've gone too far */
			ve_top = vfs_population_first(ve_curdir);
			return;
		}
	}
}

/**
 * @brief Scroll back until there is no more empty space at the bottom
 *        of the viewport or when the top of the playlist is reached.
 */
static void
gui_browser_cursor_adjust_emptybottom(void)
{
	struct vfs_ent *ve;
	int lines = gui_size_browser_height;

	for (ve = ve_top; ve != NULL; ve = vfs_population_next(ve)) {
		if (--lines <= 0)
			return;
	}

	for (; lines > 0 && ve_top != NULL; lines--)
		ve_top = vfs_population_prev(ve_top);

	if (ve_top == NULL)
		ve_top = vfs_population_first(ve_curdir);
}

void
gui_browser_init(void)
{
	const char *defdir;

	win_dirname = newwin(1, 0, gui_size_browser_dirname_top, 0);
	wbkgdset(win_dirname, COLOR_PAIR(GUI_COLOR_BAR));

	win_browser = newwin(gui_size_browser_height, 0,
	    gui_size_browser_top, 0);
	wbkgdset(win_browser, COLOR_PAIR(GUI_COLOR_BLOCK));

	/* Try to open the default, current, home or root directory */
	defdir = config_getopt("gui.browser.defaultpath");
	if (defdir[0] != '\0')
		ve_curdir = vfs_open(defdir, NULL);
	if (ve_curdir == NULL)
		ve_curdir = vfs_open_curwd();
#if 0
	if (ve_curdir == NULL)
		ve_curdir = vfs_open_homedir();
#endif
	if (ve_curdir == NULL)
		ve_curdir = vfs_open_rootdir();

	if (ve_curdir != NULL) {
		vfs_populate(ve_curdir);
		ve_top = ve_selected = vfs_population_first(ve_curdir);
	} else {
		ve_top = ve_selected = NULL;
	}

	gui_browser_refresh();
}

void
gui_browser_destroy(void)
{
	if (ve_curdir != NULL)
		vfs_close(ve_curdir);
	
	delwin(win_dirname);
	delwin(win_browser);
}

#ifdef BUILD_GUI_RESIZE
void
gui_browser_resize(void)
{
	GUI_LOCK;
	wresize(win_dirname, 1, COLS);
	mvwin(win_dirname, gui_size_browser_dirname_top, 0);
	clearok(win_dirname, TRUE);

	wresize(win_browser, gui_size_browser_height, COLS);
	mvwin(win_browser, gui_size_browser_top, 0);
	clearok(win_browser, TRUE);

	/* Make sure the resize didn't chop the last item */
	gui_browser_cursor_adjust_bottom();
	GUI_UNLOCK;

	gui_browser_refresh();
}
#endif /* BUILD_GUI_RESIZE */

/*
 * Cursor manipulation
 */

void
gui_browser_cursor_up(void)
{
	struct vfs_ent *ve;

	if (ve_selected != NULL) {
		ve = vfs_population_prev(ve_selected);
		if (ve != NULL) {
			/*
			 * Make sure the cursor doesn't leave the top of
			 * the screen
			 */
			if (ve_top == ve_selected)
				ve_top = ve;

			ve_selected = ve;
			gui_browser_refresh();
		} else {
			gui_msgbar_update(_("You are at the first file."), 0);
		}
	}
}

void
gui_browser_cursor_down(void)
{
	struct vfs_ent *ve;

	if (ve_selected != NULL) {
		ve = vfs_population_next(ve_selected);
		if (ve != NULL) {
			ve_selected = ve;
			
			/*
			 * Make sure the cursoer doesn't leave the
			 * bottom of the screen
			 */
			gui_browser_cursor_adjust_bottom();
			gui_browser_refresh();
		} else {
			gui_msgbar_update(_("You are at the last file."), 0);
		}
	}
}

void
gui_browser_cursor_pageup(void)
{
	int i;

	if (ve_top == vfs_population_first(ve_curdir))
		return;

	/* Cursor = bottom of the screen */
	ve_selected = vfs_population_next(ve_top);

	/* Adjust the viewport */
	for (i = 2; i < gui_size_browser_height && ve_top != NULL; i++)
		ve_top = vfs_population_prev(ve_top);
	if (ve_top == NULL)
		ve_top = vfs_population_first(ve_curdir);
	
	/* Original screen may have had one item */
	if (ve_selected == NULL)
		ve_selected = ve_top;
	gui_browser_refresh();
}

void
gui_browser_cursor_pagedown(void)
{
	int i;

	/* A la vi: page down minus two */
	for (i = 2; i < gui_size_browser_height && ve_top != NULL; i++)
		ve_top = vfs_population_next(ve_top);
	
	if (ve_top == NULL)
		ve_top = vfs_population_last(ve_curdir);
	
	ve_selected = ve_top;
	gui_browser_refresh();
}

void
gui_browser_cursor_top(void)
{
	ve_top = ve_selected = vfs_population_first(ve_curdir);
	gui_browser_refresh();
}

void
gui_browser_cursor_bottom(void)
{
	ve_selected = vfs_population_last(ve_curdir);
	gui_browser_cursor_adjust_bottom();
	gui_browser_refresh();
}

/*
 * Change current directory
 */

void
gui_browser_dir_parent(void)
{
	struct vfs_ent *ve;

	if (ve_curdir == NULL)
		return;

	ve = vfs_parent(ve_curdir);
	if (ve == NULL)
		return;

	if (vfs_populate(ve) != 0) {
		vfs_close(ve);
		return;
	}
	
	vfs_population_foreach(ve, ve_selected) {
		if (strcmp(ve_selected->name, ve_curdir->name) == 0)
			break;
	}

	vfs_close(ve_curdir);
	ve_curdir = ve;

	if (ve_selected == NULL)
		ve_selected = vfs_population_first(ve_curdir);

	gui_browser_cursor_adjust_middle();
	gui_browser_cursor_adjust_emptybottom();
	gui_browser_refresh();
}

void
gui_browser_dir_enter(void)
{
	struct vfs_ent* ve;

	if (ve_selected == NULL)
		return;

	ve = vfs_dup(ve_selected);
	if (vfs_populate(ve) != 0) {
		vfs_close(ve);
		return;
	}

	vfs_close(ve_curdir);
	ve_curdir = ve;

	ve_top = ve_selected = vfs_population_first(ve_curdir);
	gui_browser_refresh();
}

void
gui_browser_refresh(void)
{
	int i;
	char mark;
	struct vfs_ent *ve;

	GUI_LOCK;

	werase(win_dirname);
	if (ve_curdir != NULL)
		mvwaddstr(win_dirname, 0, 1, ve_curdir->filename);
	wrefresh(win_dirname);

	werase(win_browser);
	ve = ve_top;
	for (i = 0; i < gui_size_browser_height; i++) {
		if (ve == NULL) {
			/* We've reached the bottom of the list */
			mvwaddch(win_browser, i, 0, ' ');
			wbkgdset(win_browser, COLOR_PAIR(GUI_COLOR_BLOCK));
			wclrtobot(win_browser);
			break;
		}

		if (ve == ve_selected) {
			if (gui_input_curfocus == GUI_FOCUS_BROWSER)
				wbkgdset(win_browser,
				    COLOR_PAIR(GUI_COLOR_SELECT));
			else
				wbkgdset(win_browser,
				    COLOR_PAIR(GUI_COLOR_DESELECT));
		}

		/* Small whitespace on the left, or > when black & white */
		if ((ve == ve_selected) && !has_colors()) {
			if (gui_input_curfocus == GUI_FOCUS_BROWSER)
				mvwaddch(win_browser, i, 0, '>');
			else
				mvwaddch(win_browser, i, 0, '-');
		} else {
			mvwaddch(win_browser, i, 0, ' ');
		}

		waddstr(win_browser, ve->name);

		/* Marking character for dirs and such */
		mark = vfs_marking(ve);
		if (mark != '\0')
			waddch(win_browser, mark);

		wclrtoeol(win_browser);
		wbkgdset(win_browser, COLOR_PAIR(GUI_COLOR_BLOCK));

		ve = vfs_population_next(ve);
	}
	wrefresh(win_browser);

	GUI_UNLOCK;
}

/*
 * Playlist manipulation
 */

/**
 * @brief Silently scroll down one item in the file browser window,
 *        right after a track got added to the playlist.
 */
static void
gui_browser_playq_add_scroll(void)
{
	struct vfs_ent *ve;

	ve = vfs_population_next(ve_selected);
	if (ve != NULL) {
		ve_selected = ve;
		gui_browser_cursor_adjust_bottom();
		gui_browser_refresh();
	}
}

void
gui_browser_playq_add_tail(void)
{
	if (ve_selected != NULL) {
		PLAYQ_LOCK;
		playq_song_add_tail(ve_selected);
		PLAYQ_UNLOCK;
		gui_browser_playq_add_scroll();
	}
}

void
gui_browser_playq_add_head(void)
{
	if (ve_selected != NULL) {
		PLAYQ_LOCK;
		playq_song_add_head(ve_selected);
		PLAYQ_UNLOCK;
		gui_browser_playq_add_scroll();
	}
}

void
gui_browser_playq_add_after(void)
{
	if (ve_selected != NULL) {
		gui_playq_song_add_after(ve_selected);
		gui_browser_playq_add_scroll();
	}
}

void
gui_browser_playq_add_before(void)
{
	if (ve_selected != NULL) {
		gui_playq_song_add_before(ve_selected);
		gui_browser_playq_add_scroll();
	}
}

void
gui_browser_searchnext(void)
{
	struct vfs_ent *ve;

	if (gui_input_cursearch == NULL) {
		/* Nothing to search for */
		gui_msgbar_update(_("No search pattern."), 0);
		return;
	}

	if (ve_selected == NULL) {
		/* Directory is empty */
		gui_msgbar_update(_("There are no files."), 0);
		return;
	}

	/* Step 1: search from the next entry to the bottom */
	for (ve = vfs_population_next(ve_selected);
	    ve != NULL; ve = vfs_population_next(ve)) {
		if (strcasestr(ve->name, gui_input_cursearch) != NULL)
			break;
	}
	if (ve != NULL) {
		/* We've found an item below us */
		ve_selected = ve;

		/* Item may be below the viewport */
		gui_browser_cursor_adjust_bottom();
		gui_browser_refresh();
		return;
	}

	/* Step 2: search from beginning next entry to the bottom */
	for (ve = vfs_population_first(ve_curdir);
	    ve != vfs_population_next(ve_selected);
	    ve = vfs_population_next(ve)) {
		if (strcasestr(ve->name, gui_input_cursearch) != NULL)
			break;
	}
	if (ve != vfs_population_next(ve_selected)) {
		/* We've found an item above us */
		ve_selected = ve;

		/* Item may be above the viewport */
		gui_browser_cursor_adjust_top();
		gui_browser_refresh();

		/* Also show the wrapping message */
		gui_msgbar_update(_("Search wrapped to top."), 0);
		return;
	}

	gui_msgbar_update(_("Not found."), 0);
}
