/*
 * Copyright (c) 2006 Ed Schouten <ed@fxq.nl>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */
/**
 * @file gui/gui_statbar.c
 */

#include <string.h>
#include <unistd.h>
#include <sys/types.h>

#include <gui/gui_internal.h>
#include <trans.h>

/**
 * @brief Buffer containing a string representation of the playback time
 *        of the current song in the form of " x:xx / y:yy ".
 */
static GString *str_time;
/**
 * @brief Buffer containing a string representation of the artist and
 *        title of the current song.
 */
static GString *str_song;
/**
 * @brief Window object of the status bar at the top of the screen.
 */
static WINDOW *win_statbar;

void
gui_statbar_init(void)
{
	/* Bar at the top of the screen */
	win_statbar = newwin(1, 0, 0, 0);
	wbkgdset(win_statbar, COLOR_PAIR(GUI_COLOR_BAR));

	str_time = g_string_sized_new(24);
	str_song = g_string_sized_new(128);

	gui_statbar_update(NULL, 0);
}

void
gui_statbar_destroy(void)
{
	delwin(win_statbar);
	g_string_free(str_time, TRUE);
	g_string_free(str_song, TRUE);
}

#ifdef BUILD_GUI_RESIZE
void
gui_statbar_resize(void)
{
	GUI_LOCK;
	wresize(win_statbar, 1, COLS);
	clearok(win_statbar, TRUE);
	GUI_UNLOCK;

	gui_statbar_refresh();
}
#endif /* BUILD_GUI_RESIZE */

void
gui_statbar_refresh(void)
{
	GUI_LOCK;
	/* Blank it */
	werase(win_statbar);
	
	/* Put the content back */
	mvwaddstr(win_statbar, 0, 1, str_song->str);
	mvwaddstr(win_statbar, 0, COLS - str_time->len, str_time->str);

	/* And draw it */
	wrefresh(win_statbar);
	GUI_UNLOCK;
}

/**
 * @brief Fills the str_song with the artist and song value of the audio
 *        file.
 */
static void
gui_statbar_song(struct audio_file *fd)
{
	const char *artist, *title;

	if (fd == NULL) {
		g_string_assign(str_song, _("Idle"));
	} else {
		artist = fd->tag.artist ? fd->tag.artist :
		    _("Unknown artist");
		title = fd->tag.title ? fd->tag.title :
		    _("Unknown song");
		g_string_printf(str_song, _("Playing %s - %s"),
		    artist, title);
	}

#ifdef __FreeBSD__
	/* Print same message in ps(1) */
	setproctitle("%s", str_song->str);
#endif /* __FreeBSD__ */
}

/**
 * @brief Fills the str_time with the time value of the audio file.
 *        audio file.
 */
static void
gui_statbar_time(struct audio_file *fd)
{
	if (fd == NULL) {
		g_string_assign(str_time, " ");
	} else {
		g_string_printf(str_time, "  %u:%02u / %u:%02u ",
		    (int)fd->time_cur / 60, (int)fd->time_cur % 60,
		    (int)fd->time_len / 60, (int)fd->time_len % 60);
	}
}

void
gui_statbar_update(struct audio_file *fd, int timeonly)
{
	GUI_LOCK;
	if (!timeonly)
		gui_statbar_song(fd);
	gui_statbar_time(fd);
	GUI_UNLOCK;
	
	gui_statbar_refresh();
}
