/*
 * Copyright (c) 2006 Ed Schouten <ed@fxq.nl>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */
/**
 * @file playq.c
 */

#include <subr.h>
#include <playq.h>
#include <audio/audio.h>
#include <audio/output/output.h>
#include <gui/gui.h>

int playq_len;
struct playq_head playq_list =
    TAILQ_HEAD_INITIALIZER(playq_list);

GMutex 				*playq_lock;
/**
 * @brief Conditional variable used to kick the playq alive when it was
 *        waiting for a new song to be added to the playlist or when
 *        paused.
 */
static GCond			*playq_wakeup;

/**
 * @brief Reference to playback thread.
 */
static GThread			*playq_runner;

/* Playlist modifier commands */
static volatile int		do_pause;
static volatile int		do_skip;
static volatile int		do_quit;
static volatile double		do_seek;
/**
 * @brief Randomization seed used for playlist shuffling.
 */
static GRand			*playq_rand;

/**
 * @brief Deallocate a playlist entry.
 */
static void
playq_entry_free(struct playq_entry *ent)
{
	g_free(ent->name);
	g_free(ent->filename);
	g_slice_free(struct playq_entry, ent);
}

/**
 * @brief Allocate a playlist entry.
 */
static struct playq_entry *
playq_entry_alloc(char *filename, char *name)
{
	struct playq_entry *ent;

	ent = g_slice_new(struct playq_entry);
	ent->filename = g_strdup(filename);
	ent->name = g_strdup(name);

	return (ent);
}

/**
 * @brief Infinitely play music in the playlist, honouring the do_
 *        flags.
 */
static void *
playq_runner_thread(void *unused)
{
	struct playq_entry	*new;
	struct audio_file	*cur;

	h_thread_block_signals();

	do {
		/* Wait until there's a song available */
		PLAYQ_LOCK;
		while ((new = TAILQ_FIRST(&playq_list)) == NULL) {
			/* Change the current status to idle */
			gui_statbar_update(NULL, 0);

			g_cond_wait(playq_wakeup, playq_lock);
			if (do_quit) {
				PLAYQ_UNLOCK;
				goto done;
			}
		}
		gui_playq_notify_pre_removal(1);
		TAILQ_REMOVE(&playq_list, new, next);
		playq_len--;
		gui_playq_notify_done();
		PLAYQ_UNLOCK;

		g_assert(new != NULL);

		cur = audio_open(new->filename);
		playq_entry_free(new);
		
		/* Skip broken songs */
		if (cur == NULL)
			continue;
		gui_statbar_update(cur, 0);

		do_pause = do_skip = do_seek = 0;

		do  {
			if (do_pause) {
				/*
				 * XXX: Stupidness - you *MUST* specify
				 * a mutex when waiting for a
				 * conditional variable!
				 */
				PLAYQ_LOCK;
				while (do_pause && !do_quit)
					g_cond_wait(playq_wakeup, playq_lock);
				PLAYQ_UNLOCK;
			}

			if (do_seek != 0) {
				audio_seek(cur, do_seek);
				do_seek = 0;
			}

			gui_statbar_update(cur, 1);
		} while (!do_quit && !do_skip && (audio_output_play(cur) > 0));

		audio_close(cur);
	} while (!do_quit);
done:
	return (NULL);
}

void
playq_init(void)
{
	playq_lock = g_mutex_new();
	playq_wakeup = g_cond_new();
	playq_rand = g_rand_new();
}

void
playq_spawn(void)
{
	do_quit = 0;
	playq_runner = g_thread_create_full(playq_runner_thread, NULL,
	    0, 1, TRUE, G_THREAD_PRIORITY_URGENT, NULL);
}

void
playq_shutdown(void)
{
	do_quit = 1;
	g_cond_signal(playq_wakeup);
	g_thread_join(playq_runner);
}

/*
 * Public queue functions: playq_song_*
 */

/**
 * @brief Recursively add a VFS entity to the head of the playlist.
 */
static void
do_playq_song_add_head(struct vfs_ent *ve)
{
	struct playq_entry *ent;
	struct vfs_ent *cve;

	if (vfs_playable(ve)) {
		ent = playq_entry_alloc(ve->filename, ve->name);

		TAILQ_INSERT_HEAD(&playq_list, ent, next);
		playq_len++;
		gui_playq_notify_post_insertion(1);
	} else {
		if (vfs_populate(ve) == 0) {
			vfs_population_foreach_reverse(ve, cve) {
				if (vfs_recurse(cve))
					do_playq_song_add_head(cve);
			}
		}
	}
}

void
playq_song_add_head(struct vfs_ent *ve)
{
	do_playq_song_add_head(ve);
	g_cond_signal(playq_wakeup);
	gui_playq_notify_done();
}

/**
 * @brief Recursively add a VFS entity to the tail of the playlist.
 */
static void
do_playq_song_add_tail(struct vfs_ent *ve)
{
	struct playq_entry *ent;
	struct vfs_ent *cve;

	if (vfs_playable(ve)) {
		ent = playq_entry_alloc(ve->filename, ve->name);

		TAILQ_INSERT_TAIL(&playq_list, ent, next);
		playq_len++;
		gui_playq_notify_post_insertion(playq_len);
	} else {
		if (vfs_populate(ve) == 0) {
			vfs_population_foreach(ve, cve) {
				if (vfs_recurse(cve))
					do_playq_song_add_tail(cve);
			}
		}
	}
}

void
playq_song_add_tail(struct vfs_ent *ve)
{
	do_playq_song_add_tail(ve);
	g_cond_signal(playq_wakeup);
	gui_playq_notify_done();
}

/**
 * @brief Seek the current song by a certain amount of time.
 */
static void
playq_cursong_seek(double len)
{
	do_seek += len;
	/* We could be paused */
	do_pause = 0;
	g_cond_signal(playq_wakeup);
}

void
playq_cursong_seek_forward(void)
{
	playq_cursong_seek(5);
}

void
playq_cursong_seek_backward(void)
{
	playq_cursong_seek(-5);
}

void
playq_cursong_skip(void)
{
	do_skip = 1;
	/* We could be paused */
	do_pause = 0;
	g_cond_signal(playq_wakeup);
}

void
playq_cursong_pause(void)
{
	if (do_pause) {
		/* Kick it back alive */
		do_pause = 0;
		g_cond_signal(playq_wakeup);
	} else
		/* Pause on next read */
		do_pause = 1;
}

void
playq_song_fast_remove(struct playq_entry *pe, int index)
{
	gui_playq_notify_pre_removal(index);
	TAILQ_REMOVE(&playq_list, pe, next);
	playq_entry_free(pe);
	playq_len--;
	gui_playq_notify_done();
}

/**
 * @brief Recursively add a VFS entity before the specified entry.
 */
static void
do_playq_song_add_before(struct vfs_ent *ve, struct playq_entry *pe,
    int *idx_last)
{
	struct playq_entry *ent;
	struct vfs_ent *cve;

	if (vfs_playable(ve)) {
		ent = playq_entry_alloc(ve->filename, ve->name);

		TAILQ_INSERT_BEFORE(pe, ent, next);
		playq_len++;
		gui_playq_notify_post_insertion((*idx_last)++);
	} else {
		if (vfs_populate(ve) == 0) {
			vfs_population_foreach(ve, cve) {
				if (vfs_recurse(cve))
					do_playq_song_add_before(cve, pe,
					    idx_last);
			}
		}
	}
}

void
playq_song_fast_add_before(struct vfs_ent *ve, struct playq_entry *pe,
    int index)
{
	do_playq_song_add_before(ve, pe, &index);
	g_cond_signal(playq_wakeup);
	gui_playq_notify_done();
}

/**
 * @brief Recursively add a VFS entity after the specified entry.
 */
static void
do_playq_song_add_after(struct vfs_ent *ve, struct playq_entry *pe,
    int index)
{
	struct playq_entry *ent;
	struct vfs_ent *cve;

	if (vfs_playable(ve)) {
		ent = playq_entry_alloc(ve->filename, ve->name);

		TAILQ_INSERT_AFTER(&playq_list, pe, ent, next);
		playq_len++;
		gui_playq_notify_post_insertion(index + 1);
	} else {
		if (vfs_populate(ve) == 0) {
			vfs_population_foreach_reverse(ve, cve) {
				if (vfs_recurse(cve))
					do_playq_song_add_after(cve, pe,
					    index);
			}
		}
	}
}

void
playq_song_fast_add_after(struct vfs_ent *ve, struct playq_entry *pe,
    int index)
{
	do_playq_song_add_after(ve, pe, index);
	g_cond_signal(playq_wakeup);
	gui_playq_notify_done();
}

void
playq_song_fast_moveup(struct playq_entry *pe, int index)
{
	struct playq_entry *ppe;
	
	ppe = TAILQ_PREV(pe, playq_head, next);
	gui_playq_notify_pre_removal(index - 1);
	TAILQ_REMOVE(&playq_list, ppe, next);
	TAILQ_INSERT_AFTER(&playq_list, pe, ppe, next);
	gui_playq_notify_post_insertion(index);
	gui_playq_notify_done();
}

void
playq_song_fast_movedown(struct playq_entry *pe, int index)
{
	struct playq_entry *npe;
	
	npe = TAILQ_NEXT(pe, next);
	gui_playq_notify_pre_removal(index + 1);
	TAILQ_REMOVE(&playq_list, npe, next);
	TAILQ_INSERT_BEFORE(pe, npe, next);
	gui_playq_notify_post_insertion(index);
	gui_playq_notify_done();
}

void
playq_song_remove_all(void)
{
	struct playq_entry *pe;

	PLAYQ_LOCK;
	while (!TAILQ_EMPTY(&playq_list)) {
		gui_playq_notify_pre_removal(1);
		pe = TAILQ_FIRST(&playq_list);
		TAILQ_REMOVE(&playq_list, pe, next);
		playq_entry_free(pe);
		playq_len--;
	}
	gui_playq_notify_done();
	PLAYQ_UNLOCK;
}

void
playq_song_randomize(void)
{
	int remaining, idx;
	struct playq_entry *pe;

	PLAYQ_LOCK;
	/*
	 * This routine randomizes the playlist. It works quite simple.
	 * It just constantly picks random entries from the beginning
	 * and appends them to the end of the playlist.
	 */
	for (remaining = playq_len; remaining > 0; remaining--) {
		/* Fetch a random object */
		pe = TAILQ_FIRST(&playq_list);
		for (idx = g_random_int_range(0, remaining); idx > 0; idx--) {
			pe = TAILQ_NEXT(pe, next);
			g_assert(pe != NULL);
		}

		/* Remove it and place it at the end */
		TAILQ_REMOVE(&playq_list, pe, next);
		TAILQ_INSERT_TAIL(&playq_list, pe, next);
	}

	gui_playq_notify_post_randomization();
	gui_playq_notify_done();
	PLAYQ_UNLOCK;
}
