/*
 * Copyright (c) 2006 Martijn van den Dungen <martijn@il.fontys.nl>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */
/**
 * @file vfs/modules/vfs_dir.c
 */

#include <glib.h>

#include <stdio.h>
#include <string.h>
#include <dirent.h>

#include <config.h>
#include <vfs/vfs.h>
#include <vfs/modules/vfs_dir.h>

int
vfs_dir_open (struct vfs_ent *ve, int isdir)
{
	return (!isdir);
}

int
vfs_dir_populate (struct vfs_ent *ve)
{
	DIR *dir;
	struct dirent *cur;
	char *fn;
	struct vfs_ent *new, *scur;
	int hide_dotfiles;

	hide_dotfiles = config_getopt_bool("vfs.dir.hide_dotfiles");

#ifdef G_OS_WIN32
	if (strlen(ve->filename < 3)) {
		/*
		 * On Windows, A:\ is a different thing than A:. The
		 * first one represents the root directory of drive A:,
		 * while the latter one is the 'current working
		 * directory' of the drive A:. If our filename is too
		 * short, just append a trailing \.
		 */
		fn = g_strdup_printf("%s" G_DIR_SEPARATOR_S, ve->filename);
		dir = opendir(fn);
		g_free(fn);
	} else
#endif /* G_OS_WIN32 */
		dir = opendir(ve->filename);

	if (dir == NULL)
		return (-1);
	
	while ((cur = readdir(dir)) != NULL) {
		/* Skip '.' and '..' - infinite looping */
		if (cur->d_name[0] == '.' &&
		    (cur->d_name[1] == '\0' ||
		    (cur->d_name[1] == '.' && cur->d_name[2] == '\0')))
			continue;

		/* Hide dotted files */
		if (hide_dotfiles && cur->d_name[0] == '.')
			continue;

		/*
		 * Form a new filename by concatenating the directory
		 * and filename.
		 */
		fn = g_strdup_printf("%s" G_DIR_SEPARATOR_S "%s",
		    ve->filename, cur->d_name);
		new = vfs_open(fn, NULL);
		g_free(fn);

		if (new == NULL)
			continue;

		/*
		 * Add the items to the tailq in a sorted manner.
		 */
		if (TAILQ_EMPTY(&ve->population)) {
			TAILQ_INSERT_HEAD(&ve->population, new, next);
		} else {
			TAILQ_FOREACH(scur, &ve->population, next) {
				/* Store the file if the sorting priority is lower */
				if (vfs_sortorder(new) < vfs_sortorder(scur) ||
				    /* Or if they are the same and the filename is lower */
				    (vfs_sortorder(new) == vfs_sortorder(scur) &&
				    strcasecmp(new->name, scur->name) < 0)) {
					TAILQ_INSERT_BEFORE(scur, new, next);
					break;
				}
			}

			if (scur == NULL)
				TAILQ_INSERT_TAIL(&ve->population, new, next);
		}
	}

	closedir(dir);

	return (0);
}
