/*
 * Copyright (c) 2006 Ed Schouten <ed@fxq.nl>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */
/**
 * @file vfs/modules/vfs_playlist.c
 */

#include <glib.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <subr.h>
#include <vfs/vfs.h>
#include <vfs/modules/vfs_playlist.h>

/**
 * @brief Check if the extension of the given filename matches.
 */
static int
vfs_playlist_check_ext(char *name, char *ext)
{
	size_t fnlen, extlen;

	fnlen = strlen(name);
	extlen = strlen(ext);
	if (fnlen < extlen)
		return (-1);
	
	return strcasecmp(ext, &name[fnlen - extlen]);
}

/**
 * @brief Open a VFS entry by relative Win32 pathname and add it to the
 *        tail of the current entry.
 */
static int
vfs_playlist_add_tail(struct vfs_ent *ve, char *fn, char *title)
{
	char *newfn, *esc;
	struct vfs_ent *new = NULL;

#if defined(G_OS_UNIX)
	/* \... or /... */
	if (fn[0] == '\\' || fn[0] == G_DIR_SEPARATOR) {
#elif defined(G_OS_WIN32)
	/* X:\... */
	if (g_ascii_isalpha(fn[0]) &&
	    fn[1] == ':' && fn[2] == G_DIR_SEPARATOR) {
#endif
		/* Absolute pathname */
		newfn = esc = g_strdup(fn);
	} else {
		/* Relative pathname */
		newfn = g_malloc(strlen(ve->filename) + strlen(fn) + 1);
		strcpy(newfn, ve->filename);
			
		/* Chop the string after the last path separator */
		esc = strrchr(newfn, G_DIR_SEPARATOR);
		if (esc == NULL)
			goto done;
		/* We should escape after the separator */
		strcpy(++esc, fn);
	}

	/* Convert to proper separator */
#if G_DIR_SEPARATOR != '\\'
	for (; *esc != '\0'; esc++) {
		if (*esc == '\\')
			*esc = G_DIR_SEPARATOR;
	}
#endif

	/* Add it to the tail */
	new = vfs_open(newfn, title);
	if (new != NULL)
		TAILQ_INSERT_TAIL(&ve->population, new, next);
	
done:
	g_free(newfn);

	/* We've failed if we haven't added anything */
	return (new != NULL);
}

/*
 * PLS interface
 */

int
vfs_pls_open(struct vfs_ent *ve, int isdir)
{
	/* In order to speed up the process, we only match *.pls */
	if (isdir || vfs_playlist_check_ext(ve->name, ".pls") != 0)
		return (-1);

	return (0);
}

int
vfs_pls_populate(struct vfs_ent *ve)
{
	FILE *fp;
	char *ln, *ch;
	int idxoff, nidx;
	char **out;

	/* Current item */
	int idx = -1;
	char *fn = NULL;
	char *title = NULL;

	if ((fp = fopen(ve->filename, "rt")) == NULL)
		return (-1);

	while ((ln = h_fgetln(fp)) != NULL) {
		if (strncmp(ln, "File", 4) == 0) {
			/* Filename */
			out = &fn;
			idxoff = 4;
		} else if (strncmp(ln, "Title", 5) == 0) {
			/* Song title */
			out = &title;
			idxoff = 5;
		} else
			/* Unknown line */
			goto next;

		/* Parse the index number */
		nidx = atoi(ln + idxoff);
		if (nidx == 0)
			goto next;

		/* See if we have a value */
		ch = strchr(ln + idxoff, '=');
		if ((ch == NULL) || (ch[1] == '\0'))
			goto next;

		if (nidx != idx) {
			/* We've arrived at the next entry */
			if (fn != NULL)
				vfs_playlist_add_tail(ve, fn, title);

			g_free(fn); fn = NULL;
			g_free(title); title = NULL;
			idx = nidx;
		}

		/* Copy the new value */
		g_free(*out);
		*out = g_strdup(ch + 1);
next:
		g_free(ln);
	}

	/* Don't forget our trailing entry */
	if (fn != NULL)
		vfs_playlist_add_tail(ve, fn, title);

	g_free(fn);
	g_free(title);
	fclose(fp);
	
	return (0);
}

/*
 * M3U interface
 */

int
vfs_m3u_open(struct vfs_ent *ve, int isdir)
{
	/* In order to speed up the process, we only match *.m3u */
	if (isdir || vfs_playlist_check_ext(ve->name, ".m3u") != 0)
		return (-1);

	return (0);
}

int
vfs_m3u_populate(struct vfs_ent *ve)
{
	FILE *fp;
	char *ch;
	char *ln, *title = NULL;

	if ((fp = fopen(ve->filename, "rt")) == NULL)
		return (-1);

	while ((ln = h_fgetln(fp)) != NULL) {
		if (ln[0] == '#') {
			/* Only EXTINF is supported */
			if (strncmp(ln, "#EXTINF:", 8) == 0) {
				/* Consolidate double lines */
				g_free(title);
				title = NULL;

				/* Remove the duration in seconds */
				if (((ch = strchr(ln + 8, ',')) != NULL) &&
				    (ch[1] != '\0'))
					title = g_strdup(ch + 1);
			}
		} else if (ln[0] != '\0') {
			if (vfs_playlist_add_tail(ve, ln, title) == 0) {
				g_free(title);
				title = NULL;
			}
		}

		g_free(ln);
	}

	/* Trailing EXTINF */
	g_free(title);
	fclose(fp);

	return (0);
}
