/*
 * Copyright (c) 2006 Ed Schouten <ed@fxq.nl>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id$
 */
/**
 * @file vfs/vfs.h
 */

#ifndef _VFS_VFS_H_
#define _VFS_VFS_H_

#include <glib.h>

#include <queue.h>

struct vfs_module;
struct vfs_ent;

/**
 * @brief Module representing a type of file or directory on disk, containing
 *        functions to read its contents (populate it).
 */
struct vfs_module {
	/**
	 * @brief Attach the VFS module to a VFS entity
	 */
	int	(*v_open)(struct vfs_ent *ve, int isdir);
	/**
	 * @brief Detach the VFS module to a VFS entity
	 */
	void	(*v_close)(struct vfs_ent *ve);
	/**
	 * @brief Populate the VFS entity with its childs
	 */
	int	(*v_populate)(struct vfs_ent *ve);

	/**
	 * @brief Recurse through this object when the parent is recursed
	 */
	int	recurse;

	/**
	 * @brief Entity represents a file that is a regular media file
	 */
	int	playable;

	/**
	 * @brief Sorting number of items that should appear at the top
	 */
#define VFS_SORT_FIRST	0
	/**
	 * @brief Sorting number of items that should appear at the bottom
	 */
#define VFS_SORT_LAST	255
	/**
	 * @brief Order in which files should be sorted in the browser
	 */
	unsigned char sortorder;

	/**
	 * @brief Charachter placed behind the character marking the filetype
	 */
	char	marking;
};

/**
 * @brief A VFS entity is an object representing a single file or directory on
 *        disk. Each VFS entity is handled by some kind of VFS module. By
 *        default, a VFS entity only contains the filename and a reference
 *        count. It must be manually populated to get its actual contents.
 */
struct vfs_ent {
	/**
	 * @brief The name of the current object (the basename)
	 */
	char	*name;
	/**
	 * @brief The complete filename of the object (realpath)
	 */
	char	*filename;

	/**
	 * @brief The reference count of the current object, used to determine
	 *        whether it should be deallocated.
	 */
	int	refcount;

	/**
	 * @brief The VFS module responsible for handling the entity
	 */
	struct vfs_module *vmod;

	/**
	 * @brief References to to its siblings
	 */
	TAILQ_ENTRY(vfs_ent) next;
	/**
	 * @brief References to its children
	 */
	TAILQ_HEAD(vfs_pop, vfs_ent) population;
};

#ifdef G_OS_WIN32
#define WIN32_ROOT "Root"
#endif /* G_OS_WIN32 */

#ifdef BUILD_LOCKUP
/**
 * @brief Try to lock the application in a specified directory on
 *        startup.
 */
int	vfs_lockup(void);
#endif /* BUILD_LOCKUP */

/**
 * @brief Create a VFS entity from a filename. The name argument is
 *        optional. It only allows you to display entities with a
 *        different name (inside playlists).
 */
struct vfs_ent	*vfs_open(const char *filename, const char *name);
/**
 * @brief Open and return the parent of the entity.
 */
struct vfs_ent	*vfs_parent(struct vfs_ent *ve);
/**
 * @brief Duplicate the entity by increasing the reference count.
 */
struct vfs_ent	*vfs_dup(struct vfs_ent *ve);
/**
 * @brief Decrease the reference count of the entity and deallocate the
 *        entity when no references are left.
 */
void		vfs_close(struct vfs_ent *ve);
/**
 * @brief Populate the VFS entity with references to its children.
 */
int		vfs_populate(struct vfs_ent *ve);
/**
 * @brief Create a VFS entity of the current directory.
 */
#define vfs_open_curwd() vfs_open(".", NULL)
/**
 * @brief Create a VFS entity of the root directory.
 */
#ifdef WIN32_ROOT
#define vfs_open_rootdir() vfs_open(WIN32_ROOT, NULL)
#else /* !WIN32_ROOT */
#define vfs_open_rootdir() vfs_open("/", NULL)
#endif /* WIN32_ROOT */
/**
 * @brief Determine if a VFS entity is playable audio.
 */
#define vfs_playable(ve) ((ve)->vmod->playable)
/**
 * @brief Get the character that should be used to mark the file with in
 *        the filebrowser.
 */
#define vfs_marking(ve) ((ve)->vmod->marking)
/**
 * @brief Get the sorting priority of the current object.
 */
#define vfs_sortorder(ve) ((ve)->vmod->sortorder)
/**
 * @brief Determine if we should recurse this object.
 */
#define vfs_recurse(ve) ((ve)->vmod->recurse)

/**
 * @brief Walk backwards through an entity's population.
 */
#define vfs_population_foreach(ve,ent) \
	TAILQ_FOREACH(ent,&(ve)->population, next)
/**
 * @brief Walk forward through an entity's population.
 */
#define vfs_population_foreach_reverse(ve,ent) \
	TAILQ_FOREACH_REVERSE(ent,&(ve)->population,vfs_pop,next)

/**
 * @brief Return the first item in the population.
 */
#define vfs_population_first(ve) \
	TAILQ_FIRST(&(ve)->population)

/**
 * @brief Return the last item in the population.
 */
#define vfs_population_last(ve) \
	TAILQ_LAST(&(ve)->population,vfs_pop)

/**
 * @brief Return the next item in the population.
 */
#define vfs_population_next(ve) \
	TAILQ_NEXT(ve,next)

/**
 * @brief Return the previous item in the population.
 */
#define vfs_population_prev(ve) \
	TAILQ_PREV(ve,vfs_pop,next)

#endif /* !_VFS_VFS_H_ */
